package at.ac.uibk.gitsearch.repository.vocabulary;

import at.ac.uibk.gitsearch.domain.vocabulary.VocabularyItem;
import at.ac.uibk.gitsearch.domain.vocabulary.VocabularyItem.LanguageItem;
import com.fasterxml.jackson.databind.ObjectMapper;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.lang.reflect.Array;
import java.net.HttpURLConnection;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URLEncoder;
import java.nio.charset.StandardCharsets;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.cache.annotation.CacheEvict;
import org.springframework.cache.annotation.Cacheable;
import org.springframework.context.ApplicationContext;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Repository;

@Repository
public class VocabularyRepositoryImpl implements VocabularyRepository {

    private final Logger log = LoggerFactory.getLogger(VocabularyRepositoryImpl.class);

    private Map<String, URI> vocabularyURLMap = null;

    @Autowired
    private ApplicationContext applicationContext;

    @Value("${application.oeResource.apiLink}")
    private String oeResourceApiLink;

    public static class VocabularyRepositoryException extends RuntimeException {

        private static final long serialVersionUID = 7246650384858777011L;

        public VocabularyRepositoryException(String message, Throwable cause) {
            super(message, cause);
        }

        public VocabularyRepositoryException(String message) {
            super(message);
        }
    }

    @CacheEvict(allEntries = true, cacheNames = { "at.ac.uibk.gitsearch.repository.vocabulary.VocabularyRepository" })
    @Scheduled(fixedDelay = 3600000) // 1 hour in msecs
    public void cacheEvict() {
        // just for scheduling cache eviction
    }

    @Override
    @Cacheable(value = "at.ac.uibk.gitsearch.repository.vocabulary.VocabularyRepository", sync = true)
    public Map<String, URI> getVocabularyURLMap() {
        if (vocabularyURLMap == null) {
            loadVocabulary();
        }
        return vocabularyURLMap;
    }

    protected void loadVocabulary() {
        synchronized (this) {
            try {
                URI uri = new URI(oeResourceApiLink);
                final VocabularyItem[] vocabularyTypeMap = getVocabularyItemsFromURI(uri);
                vocabularyURLMap = new HashMap<>();
                Arrays
                    .stream(vocabularyTypeMap)
                    .forEach(vi -> {
                        String usedLanguageCode = "en"; // "en"
                        final Optional<LanguageItem> itemO = Arrays
                            .stream(vi.getLanguageItem())
                            .filter(entry -> usedLanguageCode.equals(entry.getLanguageCode()))
                            .findAny();
                        itemO.ifPresentOrElse(
                            item -> {
                                String type = item.getTitle();
                                vocabularyURLMap.put(
                                    type,
                                    uri.resolve(
                                        "./v1/" + URLEncoder.encode(type, StandardCharsets.UTF_8).replace("+", "%20") + "?format=json"
                                    )
                                );
                            },
                            () -> log.warn("Cannot find URL in {}", vi)
                        );
                    });
            } catch (URISyntaxException e) {
                log.error("Cannot read vocabulary!", e);
            }
        }
    }

    private VocabularyRepository getProxiedSelf() {
        return applicationContext.getBean(VocabularyRepository.class);
    }

    @SuppressWarnings("unchecked")
    @Override
    public <T extends VocabularyItem> T[] getVocabularyItemsFor(String subject, Class<?> type) {
        final VocabularyItem[] vocabularyItems = getVocabularyItemsFor(subject);
        T[] result = (T[]) Array.newInstance(type, vocabularyItems.length);
        for (int i = 0; i < vocabularyItems.length; i++) {
            result[i] = (T) vocabularyItems[i];
        }
        return result;
    }

    @Override
    @Cacheable(value = "at.ac.uibk.gitsearch.repository.vocabulary.VocabularyRepository.bySubjects", sync = true)
    public VocabularyItem[] getVocabularyItemsFor(String subject) {
        URI subjectURI = getProxiedSelf().getVocabularyURLMap().get(subject);

        if (subjectURI == null) {
            return new VocabularyItem[] {};
        }

        return getVocabularyItemsFromURI(subjectURI);
    }

    private VocabularyItem[] getVocabularyItemsFromURI(URI subjectURI) {
        try {
            HttpURLConnection conn = (HttpURLConnection) subjectURI.toURL().openConnection();
            conn.setRequestMethod("GET");
            conn.setRequestProperty("Accept", "application/json");

            if (conn.getResponseCode() != 200) {
                throw new VocabularyRepositoryException("Failed : HTTP error code : " + conn.getResponseCode());
            }
            try (BufferedReader br = new BufferedReader(new InputStreamReader((conn.getInputStream())))) {
                ObjectMapper objectMapper = new ObjectMapper();

                return objectMapper.readValue(br, VocabularyItem[].class);
            }
        } catch (IOException e) {
            log.error("Cannot read vocabulary!", e);
            return new VocabularyItem[] {};
        }
    }

    @Override
    @Cacheable(value = "at.ac.uibk.gitsearch.repository.vocabulary.VocabularyRepository.bySubjects", sync = true)
    public List<String> getSupportedLanguageCodes() {
        final VocabularyItem[] keywordItems = getProxiedSelf().getVocabularyItemsFor("Keyword");
        final Stream<String> languageCodeStream = Arrays
            .stream(keywordItems)
            .flatMap(kw -> Arrays.stream(kw.getLanguageItem()).map(li -> li.getLanguageCode()));
        return languageCodeStream.distinct().collect(Collectors.toList());
    }
}
